library XT_Demo;

{This is a sample X-Tension that demonstrates a little bit of everything:
- it outputs an About message
- it intercepts volume snapshot refinement and outputs some messages about files
	- it can copy files into evidence file containers
	- it can extract metadata from files
	- it can retrieve a Bitmap representation of picture of any supported file type
- it acts as a viewer X-Tension that outputs a lame message when viewing files
- it acts as a disk I/O X-Tension that manipulates the data in sector 0
}

uses Windows, XT_API;

var
	MainWnd: HWnd;
	CurrentVolume: THandle;

function XT_Init(nVersion, nFlags: DWord; hMainWnd: HWnd; lpReserved: Pointer): LongInt; stdcall; export;
begin
	if Assigned(XWF_OutputMessage) then begin
		Result:=XT_FLAG_CALLERACCEPTED_THREADSAFE;
		MainWnd:=hMainWnd;
	end else
		Result:=-1; // we do not want to be run by such a version of X-Ways Forensics that does not even export this function
end;

function XT_About(hTopWnd: HWnd; lpReserved: Pointer): LongInt; stdcall; export;
begin
	MessageBox(hTopWnd, 'This is a sample X-Tension provided by X-Ways for demonstration purposes only.', 'Demo', MB_ICONINFORMATION);
end;

procedure ASCIIToUnicode_Simple(ASCIIP: PAnsiChar; Buffer: PWideChar);
begin
	repeat
		Buffer^:=WideChar(ASCIIP^);
		Inc(Buffer);
		Inc(ASCIIP);
	until (ASCIIP-1)^=#0;
end;

function IntToPChW(i: Int64; Buffer: PWideChar): PWideChar;
var
	St: array[0..20] of AnsiChar;
begin
	Str(i, St);
	ASCIIToUnicode_Simple(St, Buffer);
	Result:=Buffer;
end;

function XT_ProcessItem(nItemID: LongInt; lpReserved: Pointer): LongInt; stdcall; export;
const
	ReadAndOutputSomeFileContent=false;
var
	DeletionStatus, Attr: DWord;
	FSize: Int64;
	nParentItemID: LongInt;
	Success: BOOL;
	hItem: THandle;
	Buf: array[0..20] of WideChar;
	LongBuf: array[0..255] of WideChar;
	Text: array[0..63] of AnsiChar;
begin
	Result:=0;

	if nItemID mod 1000=0 then
		XWF_AddComment(nItemID, 'Haha! Such a silly comment.', 2);
	//XWF_OutputMessage('XT_ProcessItem called for', 0);
	//XWF_OutputMessage(XWF_GetItemName(nItemID), $1 or $2);

	if nItemID<32 then begin
		XWF_OutputMessage('Processing', $2);

		// Retrieve deletion status
		DeletionStatus:=XWF_GetItemInformation(nItemID, XWF_ITEM_INFO_DELETION, @Success);

		// Retrieve item attributes
		Attr:=XWF_GetItemInformation(nItemID, XWF_ITEM_INFO_ATTR, @Success);

		if DeletionStatus=0 then
			XWF_OutputMessage('existing', $1 or $2)
		else
			XWF_OutputMessage('previously existing', $1 or $2);

		if Attr and $01<>0 then
			XWF_OutputMessage('directory named', $1 or $2)
		else
			XWF_OutputMessage('file named', $1 or $2);

		// Retrieve name
		XWF_OutputMessage(XWF_GetItemName(nItemID), $1 or $2);

		// Retrieve parent and its name
		nParentItemID:=XWF_GetItemParent(nItemID);
		if nParentItemID<>-1 then begin
			XWF_OutputMessage('in', $1 or $2);
			XWF_OutputMessage(XWF_GetItemName(nParentItemID), $1 or $2);
		end;

		lStrCpyW(Buf, 'ID: ');
		IntToPChW(nItemID, @Buf[4]);
		XWF_OutputMessage(Buf, $1 or $2);

		// Give these files the "already viewed" status (green color)
		XWF_SetItemInformation(nItemID, XWF_ITEM_INFO_FLAGS_SET, $00000080);

		{if XWF_GetHashSetAssocs(nItemID, LongBuf, 255)>0 then
			XWF_AddComment(nItemID, LongBuf, 2);}
	end;

	if ReadAndOutputSomeFileContent then begin
		hItem:=XWF_OpenItem(CurrentVolume, nItemID, $400{ $10}); // would not be necessary if this was XT_ProcessItemEx, just for demonstration purposes
		if hItem<>0 then begin
			FSize:=XWF_GetSize(hItem, nil);
			XWF_OutputMessage(IntToPChW(FSize, Buf), 0);

			XWF_Read(hItem, 0, Text, 32);
			XWF_OutputMessage(@Text, $1 or $2 or $4);

			XWF_Close(hItem);
		end;
	end;
end;

function XT_ProcessItemEx(nItemID: LongInt; hItem: THandle; lpReserved: Pointer): LongInt; stdcall; export;
const
	CopyFileToContainer=false;
	ShowMetadata=false;
	ProducePictureFile=false;
var
	MyCtr: THandle;
	P: Pointer;

	RII: TRasterImageInfo;
	Buf: Pointer;
	F: THandle;
	Flags, BytesWritten: DWord;
begin
	Result:=0;

	if nItemID>32 then exit;
	XWF_OutputMessage('XT_ProcessItemEx called for', 0);
	XWF_OutputMessage(XWF_GetItemName(nItemID), $1 or $2);

	if CopyFileToContainer then begin
		MyCtr:=XWF_CreateContainer('C:\Windows\Temp\Test.ctr', $102, nil);
		XWF_CopyToContainer(MyCtr, hItem, $1 or $4, 0, 0, 0, nil);
		XWF_CloseContainer(MyCtr, nil);
	end;
	if ShowMetadata and Assigned(XWF_GetMetadataEx) then begin
		Flags:=$01;
		P:=XWF_GetMetadataEx(hItem, Flags);
		if P<>nil then begin
			XWF_AddComment(nItemID, P, 0);
			XWF_ReleaseMem(P);
		end;
	end;
	if ProducePictureFile then begin
		RII.nSize:=SizeOf(TRasterImageInfo);
		RII.nItemID:=nItemID;
		RII.hItem:=hItem;
		RII.nFlags:=$01 or $02 or $04;
		RII.nWidth:=200;
		RII.nHeight:=200;
		Buf:=XWF_GetRasterImage(@RII);
		if Buf<>nil then begin
			F:=CreateFileA('Test.bmp', GENERIC_WRITE, 0, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
			WriteFile(F, Buf^, RII.nResSize, BytesWritten, nil);
			CloseHandle(F);
			VirtualFree(Buf, 0, MEM_RELEASE);
		end;
	end;
end;

var
  NoOfFilteredSearchHits, TotalNoOfSearchHits: DWord;

function XT_ProcessSearchHit(HitInfo: PSearchHitInfo): DWord; stdcall; export;
begin
	Result:=0;

	if ((HitInfo.nRelOfs>-1) and (HitInfo.nRelOfs mod 2=1))
	or ((HitInfo.nAbsOfs>-1) and (HitInfo.nAbsOfs mod 2=1)) then begin
		HitInfo.nFlags:=HitInfo.nFlags or $0008; // discard search hit
		Inc(NoOfFilteredSearchHits);
	end;

  Inc(TotalNoOfSearchHits);
end;

procedure AddEvent(hEvidence: THandle);
var
	EvtInfo: TEventInfo;
begin
	EvtInfo.nSize:=SizeOf(TEventInfo);
	EvtInfo.hEvidence:=hEvidence;
	EvtInfo.nEvtType:=0;
	EvtInfo.nFlags:=0;
	EvtInfo.Timestamp.dwLowDateTime:=1777516544; EvtInfo.Timestamp.dwHighDateTime:=30344836; //January 1, 2014
	EvtInfo.nItemID:=-1;
	EvtInfo.nOfs:=$10; //offset 16 in Disk/Partition/Volume mode, just an example
	EvtInfo.lpDescr:='Just a fake event created by a demo X-Tension, actually nothing happened here at that time.';

	if XWF_AddEvent(@EvtInfo)<>1 then
		MessageBox(MainWnd, 'Oops, X-Ways Forensics did not like this event.', 'Demo', MB_ICONHAND);
end;

function XT_Prepare(hVolume, hEvidence: THandle; nOpType: DWord; lpReserved: Pointer): LongInt; stdcall; export;
var
	EvObjTitle: array[0..MAX_PATH] of WideChar;
begin
	if nOpType in [XT_ACTION_LSS, XT_ACTION_PSS] then begin
		MessageBox(MainWnd, 'Warning! This X-Tension filters out search hits at odd offsets for demonstration purposes!', 'Demo', MB_ICONEXCLAMATION);
		NoOfFilteredSearchHits:=0;
		TotalNoOfSearchHits:=0;
		Result:=0;
	end else Result:=XT_PREPARE_CALLPI or XT_PREPARE_CALLPILATE {or XT_PREPARE_DONTOMIT} or XT_PREPARE_TARGETDIRS or XT_PREPARE_TARGETZEROBYTEFILES;

	if hEvidence<>0 then begin
		AddEvent(hEvidence);
		XWF_GetEvObjProp(hEvidence, 7, @EvObjTitle);
		XWF_OutputMessage(EvObjTitle, 0);
	end;

	CurrentVolume:=hVolume;
end;

function XT_Finalize(hVolume, hEvidence: THandle; nOpType: DWord; lpReserved: Pointer): LongInt; stdcall; export;
begin
	if (nOpType=XT_ACTION_LSS) or (nOpType=XT_ACTION_PSS) then begin
		if NoOfFilteredSearchHits > TotalNoOfSearchHits div 2 then
			XWF_OutputMessage('More than 50% of the search hits were filtered out by this X-Tension.', 0)
		else
			XWF_OutputMessage('Less than 50% of the search hits were filtered out by this X-Tension.', 0);
	end;

	Result:=0;
end;

procedure AddAnEvObjAndComputeAHashValueForItem4697;
var
	Ev, Vol, Item: THandle;
	CSW: array[0..255] of WideChar;
	HashBuf: array[0..511] of Byte;
begin
	Ev:=XWF_CreateEvObj(4,6,nil, nil);
	if Ev=0 then exit;

	Vol:=XWF_OpenEvObj(Ev, 0);
	XWF_SelectVolumeSnapshot(Vol);

	Item:=XWF_OpenItem(Vol, 4697, 0);

	PDWord(@HashBuf)^:=$01 or $10;
	PHandle(@HashBuf[4])^:=Item;
	XWF_GetHashValue(4697, @HashBuf);
end;

function XT_Done(lpReserved: Pointer): LongInt; stdcall; export;
begin
	//MessageBox(MainWnd, 'That''s it, X-Tension is done', 'Demo', MB_ICONINFORMATION);
end;

function XT_View(hItem: THandle; nItemID: LongInt; hVolume, hEvidence: THandle; lpReserved: Pointer; var nResSize: Int64): Pointer; stdcall; export;
const
	UTF16Signature: array[0..1] of Byte=($FF, $FE); //LE UTF-16
var
	MemW: PWideChar;
	MemA: PAnsiChar;
	FSize: DWord;
begin
	Result:=nil;
	nResSize:=-2;

	GetMem(MemW, 2048);
	if MemW<>nil then begin
		MemW^:=WideChar(UTF16Signature);
		lStrCpyW(MemW+1, '<html><body>');
		lStrCatW(MemW, 'Sorry, this X-Tension cannot actually help you to view files, it is just a lame demo, but just FYI this is the name of the file that you are trying to view: ');
		lStrCatW(MemW, '<b><font color="#0000FF">');
		lStrCatW(MemW, XWF_GetItemName(nItemID));
		lStrCatW(MemW, '</font></b></body></html>');
		nResSize:=lStrLenW(MemW)*2;
		Result:=MemW;
	end;

	//GetMem(MemA, 1024);
	MemA:=nil;
	if MemA<>nil then begin
		lStrCpyA(MemA, '<html><body><b><font color="#0000FF">Haha.</font></b></body></html>');
		nResSize:=lStrLenA(MemA);
		Result:=MemA;
	end;
end;

function XT_ReleaseMem(lpBuffer: Pointer): BOOL; stdcall; export;
begin
	FreeMem(lpBuffer);
end;

type
	PPrivateDiskInfo=^TPrivateDiskInfo;
	TPrivateDiskInfo=packed record
		nBytesPerSector: DWord;
		nSectorCount: Int64;
	end;

function XT_SectorIOInit(DrInfo: PDataSource): DWord; stdcall; export;
const
	SimulateFailure=false;
var
	PDI: PPrivateDiskInfo;
begin
	if SimulateFailure then begin
		Result:=0;
		lStrCpyW(DrInfo.lpPrivate, 'Unsupported data source.');
	end else begin
		Result:=XT_IOINIT_SECTORIO or XT_IOINIT_FILEIO;
		//Dec(DrInfo.nSectorCount, 16); // if for example we wish to present the data source 16 sectors smaller than it is physically

		GetMem(PDI, SizeOf(TPrivateDiskInfo));
		PDI.nBytesPerSector:=DrInfo.nBytesPerSector;
		PDI.nSectorCount:=DrInfo.nSectorCount;

		DrInfo.lpPrivate:=PDI;
	end;
end;

function XT_SectorIO(PDI: PPrivateDiskInfo; nDrive: LongInt; nSector: Int64; nCount: DWord; lpBuffer: PAnsiChar; nFlags: DWord): DWord; stdcall; export;
begin
	{Read the original sector contents and then superimpose some dummy string as a proof of concept.}
	Result:=XWF_SectorIO(nDrive, nSector, nCount, lpBuffer, @nFlags);

	if nSector=0 then
		lStrCpyA(lpBuffer+$100, '*** Disk X-Tension at work! ***')
	else
		lStrCpyA(lpBuffer+$0, '*** Disk X-Tension at work! ***');
end;

function XT_FileIO(lpPrivate: Pointer; nDrive: LongInt; hVolume, hItem: THandle; nItemID: LongInt; nOffset: Int64; lpBuffer: Pointer; nNumberOfBytes: Int64; nFlags: DWord): Int64; stdcall;
begin
	Result:=0;

	if nOffset>=0 then begin
		{Read the original file contents and then superimpose some dummy string as a proof of concept.}
		Result:=XWF_Read(hItem, nOffset, lpBuffer, nNumberOfBytes);
		if nNumberOfBytes>32 then
			lStrCpyA(lpBuffer, '**** File I/O is working! ****');
	end;
end;

function XT_SectorIODone(lpPrivate, lpReserved: Pointer): DWord; stdcall; export;
begin
	FreeMem(lpReserved);
	Result:=0;
	MessageBox(0, 'OK, enough!', 'Done', MB_ICONINFORMATION);
end;


exports
	XT_Init,
	XT_Done,
	XT_About,
  XT_Prepare,
	XT_Finalize,
	XT_ProcessItem,
	XT_ProcessItemEx,
	XT_ProcessSearchHit,

	XT_SectorIOInit,
	XT_SectorIO,
	XT_FileIO,
	XT_SectorIODone,

	XT_View,
	XT_ReleaseMem;

begin
	{Here you could insert code that gets executed when the DLL is loaded.}

end.
